"""
functionality to create the feature data necessary.
"""
from typing import List, Optional, Union
from datetime import datetime, timedelta

import pandas as pd
import statsmodels.api as sm

from ts_forecasting_pipeline.speccing import ModelSpecs
from ts_forecasting_pipeline.utils import get_closest_quarter


def construct_features(
    purpose: Union[datetime, str], specs: ModelSpecs
) -> pd.DataFrame:
    """
    Create a DataFrame based on ModelSpecs. This means loading data, potentially running a custom transformation,
    adding lags & returning only the relevant date rows.
    Purpose is either "train" or "test" (which influences for which time steps to load features),
    or a datetime object for which we need the features (useful for predicting).
    """

    if not (
        isinstance(purpose, datetime)
        or (isinstance(purpose, str) and purpose in ("train", "test"))
    ):
        raise Exception(
            "Purpose for dataframe construction needs to be either 'train', 'test' or a time_step (datetime object)."
        )

    df = pd.DataFrame()

    # load raw data series
    df[specs.outcome_var.name] = specs.outcome_var.load_series()
    for reg_spec in specs.regressors:
        df[reg_spec.name] = reg_spec.load_series()

    datetime_indices = get_time_steps(purpose, specs)

    # add lags on the outcome var
    add_lags(df, specs.outcome_var.name, specs.lags, inplace=True)

    # perform the custom transformation, if needed
    if specs.transform is not None:
        df = specs.transform(df)

    outcome_lags = [
        lag_name
        for lag_name in [
            specs.outcome_var.name + "_" + lag_to_suffix(lag) for lag in specs.lags
        ]
        if lag_name in df.columns
    ]

    # now select only relevant columns and relevant datetime indices
    relevant_columns = (
        [specs.outcome_var.name] + outcome_lags + [r.name for r in specs.regressors]
    )

    df = df[relevant_columns].loc[datetime_indices]

    # if specs.model_type == "OLS":
    #    df = sm.add_constant(df)

    return df


def get_time_steps(
    purpose: Union[str, datetime], specs: ModelSpecs
) -> pd.DatetimeIndex:
    """ get relevant datetime indices to build features for."""
    if isinstance(purpose, datetime):
        return pd.date_range(purpose, purpose, freq="15T")

    length_of_data = specs.end_of_test_data - specs.start_of_training_data
    datetime_indices = None
    if purpose == "train":
        end_of_training_data = get_closest_quarter(
            specs.start_of_training_data
            + length_of_data * specs.ratio_training_test_data
        )
        print("Start of training: %s" % specs.start_of_training_data)
        print("End of training: %s" % end_of_training_data)
        datetime_indices = pd.date_range(
            specs.start_of_training_data, end_of_training_data, freq="15T"
        )
    elif purpose == "test":
        start_of_testing_data = get_closest_quarter(
            specs.start_of_training_data
            + (length_of_data * specs.ratio_training_test_data)
            + timedelta(minutes=15)
        )
        print("Start of testing: %s" % start_of_testing_data)
        print("End of testing: %s" % specs.end_of_test_data)
        datetime_indices = pd.date_range(
            start_of_testing_data, specs.end_of_test_data, freq="15T"
        )

    return datetime_indices


def lag_to_suffix(lag: int) -> str:
    """
    Return the suffix for a column, given its lag.
    """
    if lag < 0:
        str_lag = "f" + str(abs(lag))
    else:
        str_lag = "l" + str(abs(lag))
    return str_lag


def add_lags(
    df: pd.DataFrame,
    column: str,
    lags: List[int],
    name_as: str = "quarter_hour",
    inplace: bool = True,
) -> Optional[pd.DataFrame]:
    """
    Creates lag columns for a column in the dataframe. Lags are in fifteen minute steps (15T).
    Positive values are lags, while negative values are future values.
    The new columns are named like the lagged column, plus "_l<lag>" (or "_f<lag>" for positive lags (future values)),
    where <lag> is the 15-minute lag value or the translation of it into days, weeks or years.
    In case of positive 'lags' (future values), new columns are named like the lagged column, plus "_f<lag>".

    TODO: We could also review if using statsmodels.tsa.tsatools.add_lag is of interest, but here self-made is probably
          what we need.
    """
    if inplace is False:
        df = df.copy()

    lag_names = [l for l in lags]
    if name_as == "hour":
        lag_names = [int(l / 4) for l in lags]
    if name_as == "day":
        lag_names = [int(l / 96) for l in lags]
    if name_as == "week":
        lag_names = [int(l / 96 / 7) for l in lags]
    if name_as == "year":
        lag_names = [int(l / 96 / 365) for l in lags]

    for lag in lags:
        lag_name = lag_names[lags.index(lag)]
        df[column + "_" + lag_to_suffix(lag_name)] = df[column].shift(lag)

    if inplace is False:
        return df



from typing import List, Dict, Tuple, Optional, Union
import datetime
import featuring_hendrik as featuring
import pandas as pd
from statsmodels import regression
import numpy as np
import statsmodels.api as sm
from matplotlib import pyplot as plt
import pickle
from os import listdir

"""
Functions for working with time series models.
TODO: store the exact circumstances of a model, like lags and regressors somewhere, so we can easily know how to create
fitting feature vectors. For now, we'll hardcode those.
"""

# Here we can list any class we might use. They all can be expected to have a fit() and a predict(X) method.
MODEL_CLASS = Union[regression.linear_model.OLS]

RATIO_TRAINING_TEST_DATA = 2 / 3


def load_model(outcome, model_dir):
    dir_list = listdir(model_dir)
    relevant_models = [s for s in dir_list if "model_" + outcome in s]
    model_name = sorted(relevant_models)[-1]
    print("loading model: %s" % (model_name))
    return pickle.load(open("/".join([model_dir, model_name]), "rb"))

    ## note that the variable containers include the data, loading them again in the notebook does not work right atm
    ## should be fixed asap
    ## so data frame construcor is not saved now


def save_model(model, outcome, model_dir):
    model_name = "_".join(
        [
            "model",
            outcome,
            datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M%S"),
        ]
    )
    pickle.dump(model, open(model_dir + "/" + model_name + ".sav", "wb"))


def create_model(
    df,
    outcome,
    lags,
    regressors,
    datetime_train,
    training_window,
    include_constant=True,
):
    """ 
    Parameters
    df: DataFrame
    the data

    outcome: str 
    dependent variable. variable to predict
    
    lags: list
    lags to use of outcome for prediction

    regressors: list
    independent variable(s)

    datetime: datetime
    moment up to which data will be used for fitting the model

    training_window: int
    number of days that are used for training the model
    
    NOTES: also only works with 15T freq and no missing variables
    horizon is not specifically mentioned in this function; it is implicit in the number of lags that are provided. This could be more explicit in this function

    """

    traing_start_date = datetime_train - datetime.timedelta(days=training_window)

    ## for checking purpose, hardcode traing_start_date, remove later
    # traing_start_date = datetime.datetime(2015,1,8)

    training_data_indices = pd.date_range(traing_start_date, datetime_train, freq="15T")

    outcome_lags = [
        outcome + "_" + featuring.convert_lag_to_name_str(lag) for lag in lags
    ]

    regression_frame = df[[outcome] + outcome_lags + regressors].loc[
        training_data_indices
    ]
    ## remove any observation where data is missing, other parts of the workflow cannot handle missing data, so everything should be verified
    regression_frame = regression_frame[~pd.isnull(regression_frame).any(axis=1)]
    if include_constant == True:
        regression_frame["constant"] = 1

    X_train = regression_frame.iloc[:, 1:]
    y_train = np.array(regression_frame.iloc[:, 0])

    mod = sm.OLS(y_train, X_train)
    res = mod.fit()
    return res


def plot_error_graph(true_values, predicted_values, use_abs_errors=False):

    results_df = pd.DataFrame({"y_hat_test": predicted_values, "y_test": true_values})

    ## remove 0 s
    results_df = results_df[(results_df != 0).all(1)]

    results_df["max_error"] = abs(results_df.y_hat_test / results_df.y_test - 1)
    if use_abs_errors == True:
        ## if you want to look at abs values, instead of (abs)proportional errors
        results_df["max_error"] = abs(results_df.y_hat_test - results_df.y_test)

    results_df.sort_values("max_error", inplace=True)
    results_df["proportion"] = (np.arange(len(results_df)) + 1) / len(results_df)

    plt.plot(results_df["max_error"], results_df["proportion"], "-o")
    plt.ylim(0, 1)
    plt.xlim(0, 1)
    plt.xlabel("max error for proportion")
    plt.ylabel("proportion of observations")


def test_model(
    model,
    df,
    outcome,
    lags,
    regressors,
    datetime_first_test,
    testing_window,
    include_constant=True,
    return_fitted_values=False,
):

    testing_end_date = datetime_first_test + datetime.timedelta(days=testing_window)

    ## for checking purpose, hardcode traing_start_date, remove later
    # testing_end_date = datetime.datetime(2015,4,15,3,45)

    testing_data_indices = pd.date_range(
        datetime_first_test, testing_end_date, freq="15T"
    )

    outcome_lags = [
        outcome + "_" + featuring.convert_lag_to_name_str(lag) for lag in lags
    ]

    regression_frame = df[[outcome] + outcome_lags + regressors].loc[
        testing_data_indices
    ]

    if include_constant == True:
        regression_frame["constant"] = 1

    X_test = regression_frame.iloc[:, 1:]
    y_test = np.array(regression_frame.iloc[:, 0])

    y_hat_test = model.predict(X_test)
    print(
        "rmse = %s"
        % (str(round(sm.tools.eval_measures.rmse(y_test, y_hat_test, axis=0), 4)))
    )

    plot_error_graph(y_test, y_hat_test)

    if return_fitted_values == True:
        return pd.DataFrame({"predicted_values": y_hat_test, "y_test": y_test})


def model_param_grid_search(
    df: pd.DataFrame,
    start_data: datetime,
    end_data: datetime,
    params: Dict[str, Tuple[float, float]],
) -> Dict[str, float]:

    """
    Creates and tests models with different model parameters.
    Returns the best parameter set w.r.t. smallest RMSE.
    """
    return {}



from datetime import datetime, timedelta
import pytz


def tz_aware_utc_now() -> datetime:
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def get_most_recent_quarter(dt: datetime = None) -> datetime:
    if dt is None:
        dt = tz_aware_utc_now()
    return dt.replace(minute=dt.minute - (dt.minute % 15), second=0, microsecond=0)


def get_closest_quarter(dt: datetime = None) -> datetime:
    cdt = get_most_recent_quarter(dt)
    if dt - cdt > timedelta(minutes=7, seconds=30):
        # round up
        return cdt + timedelta(minutes=15)
    return cdt


def day_lags(lags):
    """Translate day lags into 15-minute lags"""
    return [l * 96 for l in lags]



from typing import Tuple
from datetime import datetime, timedelta
from copy import deepcopy

import pandas as pd
import numpy as np
import pytz

from ts_forecasting_pipeline.speccing import ObjectSeriesSpecs
from ts_forecasting_pipeline.modelling import ModelState, create_fitted_model
from ts_forecasting_pipeline.featuring import construct_features
from ts_forecasting_pipeline.utils import get_closest_quarter


"""
Functionality for making predictions per time slot.
"""


def make_forecast_for(
    time_step: datetime, model_state: ModelState
) -> Tuple[float, ModelState]:
    """
    Make a forecast for the time slot.
    May create a model if the previous one is outdated.
    Returns a dict with horizons mapping to forecasted values, as well as the latest valid model.
    """
    if time_step.tzinfo is None:
        time_step.replace(tzinfo=pytz.utc)
    model, specs = model_state.split()
    # create model if old one is outdated.
    if time_step - specs.creation_time > specs.remodel_frequency:
        print("Fitting new model before predicting %s ..." % time_step)
        model = create_fitted_model(specs, "")
        specs.creation_time = time_step
    prediction = model.predict(
        np.hstack(construct_features(time_step, specs).values[0][1:])
    )
    print("Predicting for %s: %.2f" % (time_step, prediction))
    return prediction, ModelState(model, specs)


def make_rolling_forecasts(
    start: datetime, end: datetime, model_state: ModelState
) -> Tuple[pd.Series, ModelState]:
    """
    Repeatedly call make_forecast - for all time steps the desired time window.
    The time window of the specs (training + test data) is moved forwards also step by step.
    Will fail if series specs do not allocate enough data.
    Return Pandas.Series as result, as well as the last ModelState.
    """
    # Prepare time range
    for dt in (start, end):
        if dt.tzinfo is None:
            dt.replace(tzinfo=pytz.utc)
    start = get_closest_quarter(start)
    end = get_closest_quarter(end)

    model, specs = model_state.split()

    # load data from any source only once (i.e. reduce calls to DB or reading in pickles)
    # TODO: copy original series specs, attach them at the end with updated start and end
    # orig_outcome_var = deepcopy(specs.outcome_var)
    # orig_regressors = [deepcopy(r) for r in specs.regressors]

    specs.outcome_var = ObjectSeriesSpecs(
        name=specs.outcome_var.name, data=specs.outcome_var.load_series()
    )
    specs.regressors = [
        ObjectSeriesSpecs(name=r.name, data=r.load_series()) for r in specs.regressors
    ]

    values = pd.Series(index=pd.date_range(start, end, freq="15T"))
    time_step = start
    while time_step < end:
        value, model_state = make_forecast_for(time_step, ModelState(model, specs))
        values[time_step] = value
        # move the model specs one time step further
        specs.start_of_training_data = specs.start_of_training_data + timedelta(
            minutes=15
        )
        specs.end_of_test_data = specs.end_of_test_data + timedelta(minutes=15)
        time_step = time_step + timedelta(minutes=15)

    return values, ModelState(model, specs)



from setuptools import setup

setup(
    name="ts_forecasting_pipeline",
    description="Toolset for time series forecasting. Supports both basic modeling and in-production usage.",
    author="Seita BV",
    author_email="nicolas@seita.nl",
    keywords=["time series", "forecasting"],
    version="0.2.0",
    install_requires=[
        "pandas",
        "statsmodels",
        "matplotlib",
        "numpy",
        "pytz",
        "python-dateutil >= 2.5",
        "SQLALchemy",
    ],
    tests_require=["pytest"],
    packages=["ts_forecasting_pipeline"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    long_description="""\
    Toolset for time series forecasting, based on fundamental data science libraries like Pandas,
    statsmodels etc.
    Supports model design as well as storing and accessing forecasts.
    It contains functionality that is both useful for data & model exploration as well as 
    integrating into production code.
""",
)



"""Code to help (de)serializing ModelStates"""

import os
import json
import pickle
import copy
from datetime import date, datetime, timedelta, tzinfo


from statsmodels.base.model import LikelihoodModelResults as StatsModel
import pandas as pd

from ts_forecasting_pipeline.speccing import SeriesSpecs, DBSeriesSpecs, ModelSpecs
from ts_forecasting_pipeline import ModelState


MODEL_DIR = "models"


def save_model(model_state: ModelState, version: str):
    """
    Save (a deep copy of) model as a pickle and specs as JSON (NOTE: not thread-safe, of course.
    Best use in single-user mode).
    """
    model, specs = model_state.model, model_state.specs
    if specs.transform is not None:
        raise Exception(
            "Cannot serialise this ModelSpecs object. Transformation functions are not yet supported."
        )
    if isinstance(specs.outcome_var, DBSeriesSpecs) or any(
        isinstance(r, DBSeriesSpecs) for r in specs.regressors
    ):
        raise Exception(
            "Cannot serialise this ModelSpecs object. DBSeriesSpecs are not yet supported."
        )
    ensure_model_specs_file(MODEL_DIR)
    model_identifier = "%s_%s" % (specs.outcome_var.name, version)
    model_filename = "%s.pickle" % model_identifier
    copy.deepcopy(model).save("%s/%s" % (MODEL_DIR, model_filename), remove_data=True)
    specs.model_filename = model_filename
    with open("%s/specs.json" % MODEL_DIR, "r") as specs_file:
        existing_specs = json.loads(specs_file.read())
    existing_specs[model_identifier] = specs.as_dict()
    existing_specs_json = json.dumps(existing_specs, default=json_serial_helper)
    with open("%s/specs.json" % MODEL_DIR, "w") as specs_file:
        specs_file.write(existing_specs_json)


def load_model(outcome_var_name: str) -> ModelState:
    """
    Load a model from disk (pickle). Use outcome variable in name.
    """
    ensure_model_specs_file(MODEL_DIR)
    specs: ModelSpecs = None
    with open("%s/specs.json" % MODEL_DIR, "r") as specs_file:
        existing_specs = json.loads(specs_file.read())
        for identifier in existing_specs.keys():
            if identifier.startswith(outcome_var_name):
                specs_json = existing_specs[identifier]
                specs = ModelSpecs(**specs_json)
                break
    if specs is None:
        raise Exception("No model found which predicts %s ..." % outcome_var_name)
    model = StatsModel.load("%s/%s" % (MODEL_DIR, specs.model_filename))
    return ModelState(model, specs)


def ensure_model_specs_file(model_dir: str):
    """Make sure specs.json is in the model dir"""
    if not os.path.exists(model_dir):
        raise Exception("Cannot find MODEL_DIR: %s" % model_dir)
    filename = "%s/specs.json" % model_dir
    if not os.path.exists(filename):
        with open(filename, "w") as specs_file:
            specs_file.write("{}")


def json_serial_helper(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, (datetime, date, pd.Timestamp)):
        return obj.isoformat()
    if isinstance(obj, timedelta):
        return obj.total_seconds()
    if isinstance(obj, tzinfo):
        return str(obj)
    if isinstance(obj, SeriesSpecs):
        return json.dumps(obj.as_dict(), default=json_serial_helper)
    if isinstance(obj, pd.Series):
        return obj.to_json(date_format="iso")
    if callable(obj):
        return pickle.dumps(obj).decode(
            encoding="latin1"
        )  # attention: this dumps only the importable location!
    raise TypeError("Type %s not serializable" % type(obj))
